package brownstrategy.agent;

import java.util.Arrays;
import java.util.Random;
import java.util.StringTokenizer;

import util.Array;
import util.Constants;


import game.Game;
import game.Strategy;

public abstract class Agent implements Strategy {


	protected Game _g;
	protected String _options;
	protected long _seed;
	protected int _playerIdx;
	
	protected int _day,_cacheDay;
	protected double[] _cacheDist;


	/**
	 * Returns distribution of actions that should be probabilistically taken.
	 * Values should sum to 1.
	 * @return
	 */
	public abstract double[] getActionDistribution();	
	
	
	public abstract Agent getCopy();
	
	public abstract Agent getCopy(Game g, long seed, String options);
	
	public abstract void lookAhead(int[] actions);
	
	public abstract void lookBack();


	public Agent(Game g, long seed, String options) {
		_g = g;
		_options = options;
		_seed = seed;
		StringTokenizer st = new StringTokenizer(options, ",");
		if (st.hasMoreTokens()) {
		     _playerIdx = Integer.parseInt(st.nextToken());
		}
		else {
			_playerIdx = 0;
		}
		_day = 0;
		_cacheDay = -1;
		_cacheDist = null;
	}

	//For each action, gets the expected utility (of each player) for us playing that action,
	//Given the action distributions of the opponents.
	public double[][] getExpectedUtilitiesGivenActionDistributions(double[] p1Distribution, double[] p2Distribution) {
		double[][] expectedUtilityForEachPlayer = new double[Constants.NUM_CLOCK_POSITIONS][Constants.NUM_AGENTS];		
		for (int p0Action=0; p0Action<Constants.NUM_CLOCK_POSITIONS; p0Action++) {

			//Compute expected utilities, given we take p0Action.
			for (int p1Action=0; p1Action<Constants.NUM_CLOCK_POSITIONS; p1Action++) {
				double probP1Action = p1Distribution[p1Action];
				for (int p2Action=0; p2Action<Constants.NUM_CLOCK_POSITIONS; p2Action++) {
					double probP2Action = p2Distribution[p2Action];
					double probState = probP1Action * probP2Action;
					int[] utilityForEachPlayer = getUtilityForEachPlayer(p0Action, p1Action, p2Action);

					for (int player=0; player<Constants.NUM_AGENTS; player++) {
						expectedUtilityForEachPlayer[p0Action][player] += probState * utilityForEachPlayer[player];
					}					
				}
			}
		}
		return expectedUtilityForEachPlayer;
	}


	//For each action, gets the expected score (of each player) if we play that action, 
	//given the action distributions of the opponents. 
	//Note: "score" is just the cumulative utility of each player over the course of the game.
	public double[][] getExpectedScoresGivenActionDistributions(double[] p1Distribution, double[] p2Distribution) {
		double[][] expectedScore = new double[Constants.NUM_CLOCK_POSITIONS][Constants.NUM_AGENTS];
		double[][] expectedUtility = getExpectedUtilitiesGivenActionDistributions(p1Distribution, p2Distribution);
		for (int action=0; action<Constants.NUM_CLOCK_POSITIONS; action++) {
			for (int player=0; player<Constants.NUM_AGENTS; player++) {
				expectedScore[action][player] = expectedUtility[action][player];
			}
		}
		return expectedScore;
	}

	//Computes the distance between our score and the score of our best competitor. 
	//Positive values mean that we have the higher score than our opponent.
	public double[] getExpectedScoreDifferencesGivenActionDistributions(double[] p1Distribution, double[] p2Distribution) {
		double[] expectedScoreDiffs = new double[Constants.NUM_CLOCK_POSITIONS];

		double[][] expectedScores = getExpectedScoresGivenActionDistributions(p1Distribution, p2Distribution);
		for (int action=0; action<Constants.NUM_CLOCK_POSITIONS; action++) {
			double ourScore = expectedScores[action][0];
			double bestOpponentScore = Math.max(expectedScores[action][1], expectedScores[action][2]);
			expectedScoreDiffs[action] = ourScore - bestOpponentScore;
		}

		return expectedScoreDiffs;
	}


	//Given the following actions are played, get utility for each player
	public int[] getUtilityForEachPlayer(int p0Action, int p1Action, int p2Action) {
		return Constants.payoffs[p0Action][p1Action][p2Action];
	}


	//Get utility for player 0.
	public int getUtility(int p0Action, int p1Action, int p2Action) {
		return Constants.payoffs[p0Action][p1Action][p2Action][0];
	}
	
	//Get utility for player 0.
	public int getUtility(int[] actions) {
		return Constants.payoffs[actions[0]][actions[1]][actions[2]][0];
	}

	public Game getGame() {
		return _g;
	}
	
	public long getSeed() {
		return _seed;
	}
	
	public String getOptions() {
		return _options;
	}
	
}
