package edu.gatech.macdermed;

import game.Game;
import game.Strategy;
import game.Utility;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.Random;



/**
 * A strategy that tries to park across from someone else as quickly as possible
 * @author Liam
 *
 */
public class MatchMate implements Strategy{
	private final int NUMPLAYERS;
	private final int NUMACTIONS;
  /**
   * The action the player plays
   */
  private Random r;
  private Utility utility;  
  int myMove; 
  
  int roundNumber;
  int[] lastMoves;
  int[] moveCounts;

  enum modes { MATCH, SURRENDER };
  modes mode;
  private final int HOLD_TIME = 1;
  private final int SURRENDER_HORIZON = 12;
  private final double SURRENDER_VALUE = 6.0;
  private final int MOVING_AVG_LENGTH = 6;
  LinkedList<Integer> valueQueue;
  int queueValue;
  
  /**
   * Construct a constant strategy. The game is
   * used to get the number of actions, and the
   * seed is used to seed a pseudorandom number 
   * generator, which then is queried for the action 
   * to take.
   * 
   * @param g
   * @param seed 
   */
  public MatchMate(Game g, long seed, String options) {
	NUMPLAYERS = g.getNumPlayers();
	NUMACTIONS = g.getNumActions();
    int numActions = g.getNumActions();
    r = new Random(seed);
    myMove = r.nextInt(numActions);
    this.utility = g.getUtility();
    moveCounts = new int[NUMPLAYERS];
    Arrays.fill(moveCounts, 0);
    valueQueue = new LinkedList<Integer>();
    for(int i=0; i<MOVING_AVG_LENGTH; i++) {
    	valueQueue.add(new Integer(0));
    }
    queueValue = 0;
    mode = modes.MATCH;
    roundNumber = 0;
  }
  



  @Override
  public int getAction() {
	roundNumber ++;
    return myMove;
  }

  @Override
  public void observeOutcome(int[] actions) {
	  // record if a player has moved or not
	  if(roundNumber != 1) {
		  for(int i=0; i<NUMPLAYERS; i++) {
			  if(actions[i] != lastMoves[i])
				  moveCounts[i] ++;
		  }
	  }
	  
	  // update moving average
	  int reward = utility.getUtility(actions)[0];
	  valueQueue.add(reward);
	  queueValue += reward;
	  queueValue -= valueQueue.removeFirst();
	  
	  // if doing poorly, surrender and revert to random strategy
	  if(roundNumber >= SURRENDER_HORIZON && ((double)queueValue)/MOVING_AVG_LENGTH < SURRENDER_VALUE && ! isOppositeSomeone(actions))
		  mode = modes.SURRENDER;
	  
	  switch (mode) {
		  case MATCH:    	matchStrat(actions);  break;
		  case SURRENDER: 	myMove = r.nextInt(NUMACTIONS);  break;
		  default:	throw new Error("Bad Mode in strategy MatchMate");
	  }
	  
	  lastMoves = actions;
  }
  
  boolean isBest(int[] actions) {
	  int[] utils = utility.getUtility(actions);
	  for(int i=1; i<NUMPLAYERS; i++) {
		  if(utils[0] <= utils[i])
			  return false;
	  }
	  return true;
  }
  
  int mostSteadyOpp() {
	  int mostSteady = -1;
	  int value = Integer.MAX_VALUE;
	  int numEqual = 0;
	  for(int i=1; i<NUMPLAYERS; i++) {
		  if(moveCounts[i] < value) {
			  numEqual = 1;
			  mostSteady = i;
			  value = moveCounts[i];
		  } else if(moveCounts[i] == value) {
			  numEqual ++;
			  if(r.nextDouble() < 1.0/numEqual)
				  mostSteady = i;
		  }
	  }
	  return mostSteady;
  }
  
  public void matchStrat(int[] actions) {
	  boolean willStay = isOppositeSomeone(actions) ||
	  					 isBest(actions) ||
	  					 roundNumber <= HOLD_TIME ||
	  					 r.nextDouble() > 1.0/2.0 ;
	  if(! willStay)
		  	myMove = oppositePosition(actions[mostSteadyOpp()]);

  }
  
  boolean isOppositeSomeone(int[] actions) {
	  for(int i=1; i<NUMPLAYERS; i++)
		 if (actions[i] == oppositePosition(actions[0]))
			 return true;
	  return false;
  }
  
  int oppositePosition(int move) {
	  return (move+(NUMACTIONS/2)) % NUMACTIONS;
  }
  
}
