package game;

import java.util.Vector;

import util.Array;

/**
 * This class supports repeated games.
 * Each player considers herself to be the first player.
 * @author maz
 *
 */
public class GameDynamics {
  
  Game game;
  /**
   * The utility of the underlying game. Must be symmetric.
   */
  Utility utility;
  
  /**
   * The actions taken by the players.
   */
  int[] actions;
  
  /**
   * The actions observed by the players. Each player observes
   * his action as first.
   */
  int[][] playerBuffers;
  
  /**
   * Observers of the game.
   */
  Vector<GameObserver> observers;
  
  /**
   * The strategies of the players.
   */
  Strategy[] strategies;
  
  
  /**
   * Construct a game.
   * @param utility a utility function which must be symmetric and immutable.
   * @param strategies an array of strategies for the players.
   */
  public GameDynamics(Game g,Strategy[] strategies){
    this.utility = g.getUtility();
    this.game = g;
    this.strategies = strategies;
    actions = new int[strategies.length];
    playerBuffers = new int[strategies.length][strategies.length];
    observers=new Vector<GameObserver>();
  }
  
  
  /**
   * Add an observer. Observers see the actions played.
   * @param observer
   */
  public void addObserver(GameObserver observer){
    observers.add(observer);
  }

  
  /**
   * Play a single round of the game. Query each strategy
   * for its action, and then send the result back to all
   * strategies and observers.
   * 
   * NOTE: Each strategy views itself as the first player.
   */
  public void playOneRound(){
    for(int i=0;i<actions.length;i++){
      actions[i] = strategies[i].getAction();
      if (actions[i]>=game.getNumActions(i)){
        throw new RuntimeException("Illegal action played by "+i);
      }
      if (actions[i]<0){
        throw new RuntimeException("Illegal action played by "+i);
      }
      
    }
    
    for(int i=0;i<strategies.length;i++){
      System.arraycopy(actions, 0, playerBuffers[i], 0, actions.length);
      Array.swap(playerBuffers[i],0,i);
      strategies[i].observeOutcome(playerBuffers[i]);
    }
    
    for(GameObserver observer:observers){
      observer.observeOutcome(actions);
    }
  }
  
  /**
   * Player multiple rounds.
   * @param numIterations the number of rounds to play.
   */
  public void iterate(int numIterations){
    for(int i=0;i<numIterations;i++){
      playOneRound();
    }
  }
 
}
