package lemonade;

import java.util.Collections;
import java.util.Vector;


/**
 * A class representing the logic of positions around the island
 * for an arbitrary number of positions or players.
 * Legal positions are from
 * 0 (inclusive) to islandSize (exclusive).
 * Used by LemonadeUtility as well as HumanDisplay.
 * 
 * @author maz
 *
 */
public class IslandLogic {
  
  /**
   * The size of the island. 
   */
  int islandSize;
  
  /**
   * Construct an island of size islandSize.
   * @param islandSize
   */
  public IslandLogic(int islandSize) {
    this.islandSize = islandSize;
  }

  /**
   * Get the size of the island (number of actions)
   * @return the size of the island (number of actions)
   */
  public int getIslandSize(){
    return islandSize;
  }

  
  public int nextRight(int position){
    ++position;
    if (position==islandSize){
      return 0;
    }
    return position;
  }
  
  

  public int nextLeft(int position){
    --position;
    if (position==-1){
      return islandSize-1;
    }
    return position;
  }
  
  /**
   * Gets the distance from the player to the first player to the left
   * who is not colocated with the current player.
   * 
   * If all players are colocated, returns the circumference of the island.
   * @param positionsSorted
   * @param index the index in positionsSorted of the player of interest
   * @return the distance from the player to the first player to the left
   */
  public int getDistanceLeft(Vector<Action> positionsSorted, int index){
    
    if (index<positionsSorted.size()-1){
      int distanceLeft = positionsSorted.get(index+1).position-positionsSorted.get(index).position;
      if (distanceLeft==0){
        // Here there are points on top of one another, so ignore them.
        return getDistanceLeft(positionsSorted,index+1);
      }
      return distanceLeft;
    }
    assert(index==positionsSorted.size()-1);
    return islandSize+positionsSorted.get(0).position-
    positionsSorted.lastElement().position;
  }

  /**
   * Get the number of people on top of a certain player.
   * @param positionsSorted the positions in the order which they were sorted.
   * @param index the index of the player in the list of positions.
   * @return the number of players on top of a certain player
   */
  public int numberOfPlayersOnTop(Vector<Action> positionsSorted, int index){
    int currentPosition = positionsSorted.get(index).position;
    // The index to the left
    int firstIndex = index-1;
    while(firstIndex>=0 && positionsSorted.get(firstIndex).position==currentPosition){
      firstIndex--;
    }
    
    int lastIndex = index+1;
    while(lastIndex<positionsSorted.size() && positionsSorted.get(lastIndex).position==currentPosition){
      lastIndex++;
    }
    return (lastIndex-firstIndex)-1;
  }
  

  /**
   * Gets the distance to the player to the first person to the right (above)
   * who is not colocated with the current player.
   * 
   * If all players are colocated, returns the circumference of the island.
   * @param positionsSorted
   * @param index the index in positionsSorted of the player of interest
   * @return the distance to the player to the first person to the right (above)
   * who is not colocated with the current player
   */
  public int getDistanceRight(Vector<Action> positionsSorted, int index){
    if (index>0){
      int distanceRight = positionsSorted.get(index).position-positionsSorted.get(index-1).position;
      if (distanceRight==0){
        // Here there are points on top of one another, so ignore them.
        return getDistanceRight(positionsSorted,index-1);
      }
      return distanceRight;
    }
    assert(index==0);
    return islandSize+positionsSorted.get(0).position-
    positionsSorted.lastElement().position;
  }
  
  /**
   * Sort the positions from lowest to highest.
   * @param positions
   * @return a vector of player-position pairs sorted by position.
   */
  public Vector<Action> getPositionsSorted(int[] positions){
    Vector<Action> positionsSorted = new Vector<Action>();
    
    for(int i=0;i<positions.length;i++){
      Action a = new Action(i,positions[i]);
      positionsSorted.add(a);
    }
    Collections.sort(positionsSorted);
    return positionsSorted;    
  }
  
}
