package main;

import game.Game;
import game.Strategy;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Random;

import stats.SeriesObserver;
import stats.UtilitySeriesObserver;
import util.Array;
import lemonade.ScaledLemonadeUtility;


/**
 * A main class for running the Lemonade tournament.
 * Every triple of players is played.
 * No variance reduction tricks are used, nor is variance calculated.
 * Usage: <SEED> <PLAYER1> <OPTION1> <PLAYER2> <OPTION2> <PLAYER3> <OPTION3>
 * @author maz
 *
 */
public class RunTiming {

  
  public static SeriesObserver getObserver(Game g, String className) throws ClassNotFoundException, NoSuchMethodException, InstantiationException, InvocationTargetException, IllegalArgumentException, IllegalAccessException{
    String paramString = "(game.Game)";
    Class<?>[] params = new Class<?>[]{
        game.Game.class};
    try{
        Class<?> c = Class.forName(className);
        Constructor<?> constr = c.getConstructor(params);
        return (SeriesObserver)(constr.newInstance(g));
    } catch(ClassNotFoundException cnf){
      System.err.println("Could not find "+className);
      throw cnf;
    } catch(SecurityException s){
      System.err.println("No permission to access "+className+paramString);
      throw s;
    } catch (NoSuchMethodException nsm){
      System.err.println("Could not find "+className+paramString);
      throw nsm;
    } catch (InstantiationException ie){
      System.err.println("Cannot instantiate "+className);
      throw ie;
    } catch (InvocationTargetException ite){
      System.err.println("Constructor "+className+paramString+" threw an exception");
      throw ite;
    }
    
  }
  
  /**
   * Run a tournament with a symmetric game.
   * Every triple of players is played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public static void runTimeTest(Game g, String[] strategies, String[] options, Random seedGenerator,int iterations,int repeats, String observerClass) throws IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException{
    for(int i=0;i<strategies.length;i++){
      String[] threeStrategies = new String[]{"","strategy.UniformStrategy","strategy.UniformStrategy"};
      String[] threeOptions = new String[]{"","",""};
      threeStrategies[0]=strategies[i];
      threeOptions[0]=options[i];
      SeriesObserver observer = getObserver(g,observerClass);
      long initial = System.currentTimeMillis();
      RunLemonadeGame.manyRuns(g, threeStrategies, threeOptions, seedGenerator,iterations,repeats,observer);
      long finalTime = System.currentTimeMillis();
      double seconds = ((double)(finalTime-initial))/1000.0;
      double average = seconds/repeats;
      String passed = (average<10.0) ? "passed" : "failed";
      System.out.println(strategies[i]+" "+passed +" test. Time:"+average);

    }
  }


  public static void showUsage(){
    System.err.println("Usage:RunLemonadeTournament [-seed <SEED>|-numActions <ACTIONS>|-numRounds<ROUNDS> |-numRepeats <REPEATS>]* -players [<CLASS> <OPTION>]*");
    System.exit(0);
  }

  /**
   * Run a tournament with a symmetric game.
   * Every triple of players is played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public static void main(String[] args) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException{
    
    if (args.length<4){
      showUsage();
    }
    
    long seed=0;
    int numActions=12;
    int iterations = 100;
    String observerClassName = UtilitySeriesObserver.class.getName();
    int playerArgIndex=args.length;
    int repeats = 10;
    for(int i=0;i<args.length;i+=2){
      String arg = args[i];
      String fieldStr = args[i+1];
      if (arg.equals("-seed")){
        long field = Long.parseLong(fieldStr);
        seed = field;
      }else if (arg.equals("-actions")){
        long field = Long.parseLong(fieldStr);
        if (100L<field){
          System.err.println("Number of actions needs to be less than 100");
          showUsage();
        }
        numActions = (int)field;
      } else if (arg.equals("-rounds")){
        long field = Long.parseLong(fieldStr);
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of rounds needs to be less than 2^31-1");
          showUsage();
        }
        iterations = (int)field;
      } else if (arg.equals("-repeats")){
        long field = Long.parseLong(fieldStr);
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of repeats needs to be less than 2^31-1");
          showUsage();
        }
        repeats = (int)field;
      } else if (arg.equals("-players")){
        playerArgIndex=i+1;
        break;
      } else {
        System.err.println("Unknown option:"+arg);
        showUsage();
        
      }
    }

    Random seedGenerator = new Random(seed);
    ScaledLemonadeUtility utility = new ScaledLemonadeUtility(numActions);
    Game g = new Game(new int[]{numActions,numActions,numActions},utility);

    String[] strategies = new String[(args.length-playerArgIndex)/2];
    String[] options = new String[strategies.length];
    int playerIndex = 0;
    for(int i=playerArgIndex;i<args.length;i+=2){
      strategies[playerIndex]=args[i];
      options[playerIndex]=args[i+1];
      playerIndex++;
    }
    
    runTimeTest(g, strategies, options, seedGenerator,iterations,repeats,observerClassName);
    
  }
}
